﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Kinect;

namespace Kinect_ARsensing
{
    class NTL_EstimationCircle
    {
        private int nSample; // FIFO段数
        private int wPtr;    // FIFO書込みポインタ
        public float[] Px;   // FIFOデータ スクリーン座標x
        public float[] Py;   // FIFOデータ スクリーン座標y
        private float nSample1; // 半径の計算で使用　1/nSample
        private int rPtr;       // FIFO対応値読出しポインタ(FIFO段数の1/2だけ位置がずれている)
        private float[] wa;     // aj (Storage for Estimation)
        private float[] wb;     // bj (Storage for Estimation)
        private float[] wc;     // cj (Storage for Estimation)
        private float[] wKa;    // Kaj (Storage for Estimation)
        private float[] wKb;    // Kbj (Storage for Estimation)
        private float[] wKab1;  // Kbj (Storage for Estimation)
        private float[] wKab2;  // Kbj (Storage for Estimation)
        private float Ka;       // Ka=Σa/(a*a+b*b) ⇒ Ka=1/{Σa/(a*a+b*b)} (Storage for Estimation)
        private float Kb;       // Kb=Σb/(a*a+b*b) ⇒ Kb=1/{Σb/(a*a+b*b)} (Storage for Estimation)
        private float eXc;      // Xc (estimated parameter)
        private float eYc;      // Yc (estimated parameter)
        private float er;       // R  (estimated parameter)
        private bool FIFOfull;  // FIFOfull flag

        private float Xmin; //
        private float Xmax; //
        private float Ymin; //
        private float Ymax; //
        private NTL_FigureType FigureType; //


        public struct NTL_CircleData
        {
           public float X;
           public float Y;
           public float Radius;
           public float Xminimum;
           public float Xmaximum;
           public float Yminimum;
           public float Ymaximum;
           public NTL_FigureType FigureType;
        }


        public enum NTL_FigureType //【図形タイプを列挙型で記述】
        {
            Unknown,
            Point,
            Circle,
            Arc,
            Line
        }

        private NTL_CircleData Circle1;
       
        //*************************************************************
        //【Initialize】
        //*************************************************************
        public void Initialize(int n)
        {
            nSample = n;
            nSample1 = 1.0f / (float)n;
            wPtr = 0;
            rPtr = nSample / 2;
            Px = new float[nSample];
            Py = new float[nSample];
            wa = new float[nSample];
            wb = new float[nSample];
            wc = new float[nSample];
            wKa = new float[nSample];
            wKb = new float[nSample];
            wKab1 = new float[nSample];
            wKab2 = new float[nSample];
            eXc = 0.0f; //仮の初期値
            eYc = 0.0f; //仮の初期値
            er =0.0f;
            FIFOfull = false;
        }

        //*************************************************************
        //【Estimation】
        //*************************************************************
        public void Estimation(float x1, float y1)
        {
            //【新規スクリーン座標出現で計算するべき部分】
            float x2,y2,a,b,a2,b2,Kab;
            Px[wPtr]=x1; //【今回の骨格座標のスクリーン座標変換値をFIFOに保存する】
            Py[wPtr]=y1; //
            x2=Px[rPtr];
            y2=Py[rPtr];
            wa[wPtr]=a=2.0f*(x2-x1);
            wb[wPtr]=b=2.0f*(y2-y1);
            wc[wPtr]=x1*x1+y1*y1-x2*x2-y2*y2;
            a2=a*a;
            b2=b*b;
            if ((a2 + b2) == 0f) { Kab = 9999.0f; } else { Kab = 1.0f / (a2 + b2); } // 0div errorの対策が必要だろう★★★★★
            wKab1[wPtr]=a2*Kab; // a^2/(a^2+b^2)
            wKab2[wPtr]=b2*Kab; // b^2/(a^2+b^2)
            wKa[wPtr]=a*Kab;
            wKb[wPtr]=b*Kab;

            //【書込みポインタの更新】
            wPtr++; if(wPtr == nSample) { wPtr = 0; FIFOfull = true; }
            rPtr++; if(rPtr==nSample) { rPtr=0; }

            //【推定処理の準備】【領域も求める】
            PreEstimation();

            //【Base Transition Ruleで中心座標(Xc,Yc)を推定する】５回実行
            CircleEstimation();
            CircleEstimation();
            CircleEstimation();
            CircleEstimation();
            CircleEstimation();

            //【推定した中心座標(Xc,Yc)から半径を計算】
            CalculateRadius();

            //【軌跡が示す図形のタイプを判定】
            CheckFigureType();
        }

        //*************************************************************
        //【PreEstimation】Base Transition Ruleによる推定処理中変化しない値を計算する
        //*************************************************************
        private void PreEstimation()
        {
            int i; float sumX=0.0f, sumY=0.0f, x, y;
            Ka=0.0f;
            Kb=0.0f;
            Xmin = 10000000.0f;
            Xmax = -1.0f;
            Ymin = 10000000.0f;
            Ymax = -1.0f;
            for (i = 0; i < nSample; i++)
            {
                x = Px[i]; y = Py[i];
                sumX += x; sumY += y;
                if (Xmin > x) { Xmin = x; } //この最大、最小は値の範囲を調べておき、
                if (Xmax < x) { Xmax = x; } //推定結果の円の中心座標がこの領域外なら
                if (Ymin > y) { Ymin = y; } //円弧または直線、領域内なら円と認識する
                if (Ymax < y) { Ymax = y; } //ために使う。
               Ka+=wKab1[i];
               Kb+=wKab2[i];
            }
            Ka = 1.0f / Ka; // 0div errorの対策が必要だろう
            Kb = 1.0f / Kb; // 0div errorの対策が必要だろう
            eXc = sumX * nSample1;
            eYc = sumY * nSample1;
        }

        //*************************************************************
        //【CircleEstimation】☆☆☆☆☆ Base Transition Rule ☆☆☆☆☆
        //*************************************************************
        private void CircleEstimation()
        {
            int j; float wSa, wSb, Oj;// (working parameter) wSa=ΣajOj/(aj*aj+bj*bj), wSb=ΣbjOj/(aj*aj+bj*bj)
            //【Estimate Xc】
            wSa = 0;
            for (j = 0; j < nSample; j++)
            {
                Oj = wa[j] * eXc + wb[j] * eYc + wc[j];
                wSa += wKa[j] * Oj;
            }
            eXc -= wSa * Ka;
            //【Estimate Yc】
            wSb = 0;
            for (j = 0; j < nSample; j++)
            {
                Oj = wa[j] * eXc + wb[j] * eYc + wc[j];
                wSb += wKb[j] * Oj;
            }
            eYc -= wSb * Kb;
        }

        //*************************************************************
        //【CalculateRadius】☆☆☆☆☆ Average ☆☆☆☆☆
        //*************************************************************
        private void CalculateRadius()
        {
            float x,y;
            er = 0.0f;
            for (int i = 0; i < nSample; i++)
            {
                x=Px[i] - eXc;
                y=Py[i] - eYc;
                er += (float)Math.Sqrt(x * x + y * y); //ここはHash Table技法を使えばまだ速度改善できる。
            }
            er = er * nSample1;
        }

        //*************************************************************
        //【CheckFigureType】軌跡の図形のタイプを判定する
        //*************************************************************
        private void CheckFigureType()
        {
            const float R1 = 30.0f;  //【スクリーン座標で半径が30画素未満ならPoint】
            const float R2 = 150.0f; //【スクリーン座標で半径が150画素以上ならLine】
            FigureType = NTL_FigureType.Unknown;
            if (er < R1) { FigureType = NTL_FigureType.Point; }
            else
            {
                //【中心座標は領域の内部か？】yes⇒円, no⇒円弧or線分
                if ((eXc > Xmin) && (eXc < Xmax) && (eYc > Ymin) && (eYc < Xmax)) { FigureType = NTL_FigureType.Circle; }
                else
                {   //推定円の半径が1.0m以上なら線分。それ以外は円弧。
                    if (er > R2) { FigureType = NTL_FigureType.Line; } else { FigureType = NTL_FigureType.Arc; }
                }
            }
        }

        //*************************************************************
        //【Property】☆☆☆☆☆ Estimated Value ☆☆☆☆☆
        //*************************************************************
        public int Length  { get { return nSample; } }
        public int Pointer { get { return wPtr; } }
        public bool Full   { get { return FIFOfull; } }
        public float X { get { return eXc; } }
        public float Y { get { return eYc; } }
        public float Radius { get { return er; } }
        public float Xminimum { get { return Xmin; } }
        public float Xmaximum { get { return Xmax; } }
        public float Yminimum { get { return Ymin; } }
        public float Ymaximum { get { return Ymax; } }
        public NTL_FigureType Type { get { return FigureType; } }
        public NTL_CircleData CircleData
        {
            get
            {
                Circle1.X = eXc;
                Circle1.Y = eYc;
                Circle1.Radius = er;
                Circle1.Xminimum = Xmin;
                Circle1.Xmaximum = Xmax;
                Circle1.Yminimum = Xmin;
                Circle1.Ymaximum = Xmax;
                Circle1.FigureType = FigureType;
                return Circle1;
            }
        }
    }
}
